"""
Module with utility functions required for the plotting.
"""
from typing import Dict

import matplotlib.pyplot as plt

from .. import data_utils


def _fmt_t_str_plot(t_str):
    """Function returns a formatted string which can be plotted inside figures."""
    t_s, d_places = data_utils.get_time_sec(t_str, return_num_decimals=True)
    return rf"${t_s:.{d_places}f}\,\mathrm{{s}}$"


def _get_key_to_value_i(i: int, key_to_value: Dict):
    """Function gets the key_to_value at position i.

    Parameters
    ----------
    i: int
        Index from which to take the key_to_value.
    key_to_value: Dict
        Dictionary with keys and values, which can either be sequences or a single value

    Example
    -------
    >>> key_to_value = {"color": ["r", "b", "g"], "linestyle": ["-", "--", "-."],
                        "linewidth": 1}
    >>> _get_key_to_value_i(1, key_to_value)
    {"color": "b", "linestyle": "--", "linewidth": 1}
    """

    # Check the length of the values
    # They can be either 1 or one other value, for example either 1 or 3, but it cannot
    # be 1, 3 and 5
    value_len = set()
    for key, value in key_to_value.items():
        if hasattr(value, "__len__"):
            value_len.add(len(value))
        else:
            value_len.add(1)

    if len(value_len) > 2:
        raise ValueError(
            f"`key_to_value` is not valid, it contains {len(value_len)} "
            "different sequence lengths."
        )

    key_to_value_i = dict()
    for key, value in key_to_value.items():
        if hasattr(value, "__len__"):
            if len(value) > 1:
                key_to_value_i[key] = value[i]
            else:
                key_to_value_i[key] = value[0]
        else:
            key_to_value_i[key] = value

    return key_to_value_i


def _default_colors():
    return plt.rcParams["axes.prop_cycle"].by_key()["color"]


def _gridlib_colors():
    return ["#fe9901", "#63cabc", "#382134", "#007972", "#2333be", "#aaaaaa"]



from typing import Tuple, Dict, Union, List
import math

import numpy as np

from . import compute
from . import data_utils
from gridlib.fit import fit_grid

# TODO: add docstring
def _sample_data(data, perc: float = 0.8, seed=None):
    """_summary_

    Parameters
    ----------
    data : _type_
        _description_
    perc : float, optional
        _description_, by default 0.8
    seed : {None, int, array_like[ints], SeedSequence, BitGenerator, Generator}, optional
        A seed to initialize the BitGenerator. If None, then fresh, unpredictable
        entropy will be pulled from the OS. If an int or array_like[ints] is passed,
        then it will be passed to SeedSequence to derive the initial BitGenerator state.
        One may also pass in a SeedSequence instance. Additionally, when passed a
        BitGenerator, it will be wrapped by Generator. (THIS IS FROM NUMPY DOC, MAYBE CHANGE THIS), by default None

    Returns
    -------
    _type_
        _description_
    """
    rng = np.random.default_rng(seed)

    data_sampled = dict()
    for t_tl in data.keys():

        # Retrieve the time in seconds
        t_tl_s = data_utils.get_time_sec(t_tl)

        # The total number of data points for a time-lapse condition in the complete
        # dataset
        data_point_all = data[t_tl]["value"][0]

        # Determine how many data points should be in the resampled version
        # math.ceil returns an integer
        num_sample = math.ceil(data_point_all * perc)

        # Round the possible track lifes to the closest integer
        track_lifes = np.rint(data[t_tl]["time"] / t_tl_s)
        # Cast it to the integer dtype, since the compute survival functions needs
        # an array with dtype integer.
        track_lifes = track_lifes.astype(np.int64)

        # TODO: clean this up
        # Calculate the probability for every track life
        # ----------------------------------------------
        # First retrieve the true values and invert these, so the counts go from
        # low to high
        temp = data[t_tl]["value"][::-1]

        # Calculate the difference between every track life step
        diff = temp[1:] - temp[:-1]  # check out np.diff?
        # invert the values back so they align again with the track lifes
        diff = diff[::-1]

        # Concatenate the last value, because it equals "temp[0] - 0 = temp[0]"
        values = np.concatenate(
            (diff, np.array([temp[0]]))
        )  # add the last element again
        p = values / np.sum(values)
        # THE NEXT LINE IS WRONG!! THIS IS THE CUMULATIVE SUM, NOT THE INDIVIDUAL COUNTS
        # p = data[t_tl]["value"] / np.sum(data[t_tl]["value"])

        track_lifes_sampled = rng.choice(
            track_lifes,
            size=num_sample,
            replace=True,
            p=p,
        )

        time, value = compute.compute_survival_function(track_lifes_sampled, t_tl_s)

        # Save the results to the appropriate keys
        data_sampled[t_tl] = dict()
        data_sampled[t_tl]["time"] = time
        data_sampled[t_tl]["value"] = value

    # Returns non-normalized survival function
    return data_sampled


def resampling_grid(
    parameters,
    data,
    n: int = 10,
    perc: float = 0.8,
    seed=None,
) -> Tuple[
    Dict[str, Union[np.ndarray, float]], List[Dict[str, Union[np.ndarray, float]]]
]:
    """_summary_

    Parameters
    ----------
    parameters : _type_
        _description_
    data : _type_
        _description_
    n : int, optional
        Number of times the data should be resampled and should be fitted with GRID, by default 10.
    perc : float, optional
        The percentage of the data that should be resampled, by default 0.8
    seed : {None, int, array_like[ints], SeedSequence, BitGenerator, Generator}, optional
        A seed to initialize the BitGenerator. If None, then fresh, unpredictable
        entropy will be pulled from the OS. If an int or array_like[ints] is passed,
        then it will be passed to SeedSequence to derive the initial BitGenerator state.
        One may also pass in a SeedSequence instance. Additionally, when passed a
        BitGenerator, it will be wrapped by Generator. (THIS IS FROM NUMPY DOC, MAYBE CHANGE THIS)
    Returns
    -------
    _type_
        _description_
    """

    # Determine the full GRID fit results based on all the data.
    fit_results_all = fit_grid(parameters, data)

    rng = np.random.default_rng(seed)

    # Determine the GRID fit results for the resampled data
    fit_results_resampled = []
    for _ in range(n):

        data_resampled = _sample_data(data, perc=perc, seed=rng)

        fit_results_temp = fit_grid(parameters, data_resampled)

        fit_results_resampled.append(fit_results_temp)

    return fit_results_all, fit_results_resampled



# Here are utils function, such as for checking the whether the parameters are
#  valid
# and for reformatting the t_tls arguments
import copy
import math
import re
from typing import Dict

import numpy as np


def _num_decimal_places(value: str) -> int:
    """Function finds the number of decimal places in a string float.

    Parameters
    ----------
    value : str
        A float in string format for which the number of decimals should be determined.

    Returns
    -------
    int
        Number of decimals

    Examples
    --------
    >>> num_decimal_places("5.32")
    2
    >>> num_decimal_places("2.")
    0
    >>> num_decimal_places("98.8572")
    4
    >>> num_decimal_places("3.4120")
    3
    >>> num_decimal_places("11")
    0
    >>> num_decimal_places("0.05")
    2
    """
    # Regex from:
    # https://stackoverflow.com/questions/28749177/how-to-get-number-of-decimal-places
    # m = re.match(r"^[0-9]*\.([1-9]([0-9]*[1-9])?)0*$", value)
    # Regex is adapted such that "0.05" also gets 2 decimal places, regex above
    # requires the first value after the dot being [1-9]
    m = re.match(r"^[0-9]*\.([0-9]*[1-9])0*$", value)
    return len(m.group(1)) if m is not None else 0


def get_time_sec(t_str: str, return_num_decimals: bool = False) -> float:
    """Function retrieves the time in seconds from a time string.

    Parameters
    ----------
    t_str : str
        A value string from which the time in units of seconds needs to be retrieved.
    return_num_decimals : bool, optional
        Specifies whether the number of decimals needs to be returned. (default False)

    Returns
    -------
    t_s : float
        Time in seconds
    d_places : int, optional
        Number of decimal places. Only provided if `return_num_decimals` is True.

    Examples
    --------
    >>> get_time_sec("50ms")
    0.05
    >>> get_time_sec("4.2s")
    4.2
    >>> get_time_sec("1150ms")
    1.15
    """

    if not isinstance(t_str, str):
        raise ValueError("Input argument should be a string.")

    # pattern to match time pattern in the files, eg. 100ms or 2s
    time_pattern = re.compile(
        r"[+-]?([0-9]+([.][0-9]*)?|[.][0-9]+)", flags=re.IGNORECASE
    )
    unit_pattern = re.compile("(ms|s)", flags=re.IGNORECASE)

    # Match time and unit pattern
    m_time = time_pattern.search(t_str)
    m_unit = unit_pattern.search(t_str)

    if m_time is None or m_unit is None:
        raise ValueError("t_str is not the correct pattern")

    # get the number value of the match, eg. 50, 100, 390, 4
    t_val_str = m_time.group(0)
    t_val = float(t_val_str)

    # get the unit value of the match, eg. ms or s
    t_unit = m_unit.group(0)

    # Calculate the frame cycle time
    if t_unit == "ms":
        t_s = t_val / 1000.0
        t_s = round(t_s, len(t_val_str))

        # Store it now as a value in the seconds and now find the number of decimal
        # positions when it is stored as seconds
        t_str = f"{t_s:f}s"
        t_s, d_places = get_time_sec(t_str, return_num_decimals=True)

    elif t_unit == "s":
        t_s = t_val
        d_places = _num_decimal_places(t_val_str)
        t_s = round(t_s, d_places)

    if return_num_decimals:
        return t_s, d_places
    else:
        return t_s


def _fmt_t_str_key(t_str: str) -> str:
    t_s, d_places = get_time_sec(t_str, return_num_decimals=True)
    return f"{t_s:.{d_places}f}s"


def fmt_t_str_data(
    data: Dict[str, Dict[str, np.ndarray]]
) -> Dict[str, Dict[str, np.ndarray]]:
    """Function formats all the time-lapse times in the data dictionary to strings
    where the time-lapse times are in seconds. This allows for easier comparison and
    general processing.

    Parameters
    ----------
    data : Dict[str, Dict[str, np.ndarray]]
        Survival function data for time-lapse conditions with the following data
        structure:
            {
                "t_tl": {
                    "time": np.ndarray with the time points,
                    "value": np.ndarray with the survival function values,
                },
                ...
            }

    Returns
    -------
    Dict[str, Dict[str, np.ndarray]]
        Data dictionary but with the time-lapse time keys all in seconds instead of
        seconds or milliseconds.

    Examples
    --------
    >>> data = {"50ms": {"time": array([0.05, 0.1, 0.15]), "value": array([41, 23, 8])},
                "1s": {"time": array([1.0, 2.0]), "value": array([34, 9])}}
    >>> fmt_t_str_data(data)
    {"0.05s": {"time": array([0.05, 0.1, 0.15]), "value": array([41, 23, 8])},
     "1s": {"time": array([1.0, 2.0]), "value": array([34, 9])}}
    """
    # Deepcopy is required, because otherwise it will store just the reference pointer
    # but not the array itself
    return {_fmt_t_str_key(key): copy.deepcopy(value) for key, value in data.items()}


def process_data(
    data: Dict[str, Dict[str, np.ndarray]], delete: bool = True
) -> Dict[str, Dict[str, np.ndarray]]:
    """Function processes data so functions can be fitted to it. The processing consists
    of three steps. First, the keys of the data are changed to strings with
    the unit being seconds. The second step is optional, but the first data point of
    the shortest time-lapse time is deleted if the first time point is equal to the
    time-lapse time. Finally, the values are normalized so they represent
    probabilities.

    Parameters
    ----------
    data : Dict[str, Dict[str, np.ndarray]]
        Survival function data for time-lapse conditions with the following data
        structure:
            {
                "t_tl": {
                    "time": np.ndarray with the time points,
                    "value": np.ndarray with the survival function values,
                },
                ...
            }
    delete : bool, optional
        Flag to indicate whether the first data point of the shortest time-lapse time
        should be deleted if the first time point is equal to the time-lapse time
        (default True).

    Returns
    -------
    data_processed : Dict[str, Dict[str, np.ndarray]]
        The processed survival function data for time-lapse conditions with the
        following data structure:
            {
                "t_tl": {
                    "time": np.ndarray with the time points,
                    "value": np.ndarray with the survival function values,
                },
                ...
            }

    Examples
    --------
    >>> data = {"50ms": {"time": array([0.05, 0.1, 0.15])), "value": array([41, 23, 8])},
                "1s": {"time": array([1.0, 2.0]), "value": array([34, 9])}}
    >>> process_data(data)
    {"0.05s": {"time": array([0.1, 0.15]), "value": array([1.0, 0.34782609])},
     "1s": {"time": array([1.0, 2.0]), "value": array([1.0, 0.26470588])}}

    If `delete` is set to False:
    >>> data = {"50ms": {"time": array([0.05, 0.1, 0.15]), "value": array([41, 23, 8])},
                "1s": {"time": array([1.0, 2.0]), "value": array([34, 9])}}
    >>> process_data(data, delete=False)
    {"0.05s": {"time": array([0.05, 0.1, 0.15]), "value": [1.0, 0.56097561, 0.19512195])},
    "1s": {"time": array([1.0, 2.0]), "value": array([1.0, 0.26470588])}}
    """

    # Format the data such that all the keys are in seconds
    data_processed = fmt_t_str_data(data)

    if delete:
        t_tls = list(data_processed.keys())
        t_tls.sort()  # Sort the key values from low to high

        t_tl_min = t_tls[0]
        s_min = get_time_sec(t_tl_min)
        if math.isclose(s_min, data_processed[t_tl_min]["time"][0]):
            data_processed[t_tl_min]["time"] = data_processed[t_tl_min]["time"][1:]
            data_processed[t_tl_min]["value"] = data_processed[t_tl_min]["value"][1:]
            print(
                "WARNING: The first data point was deleted! See documentation to ",
                "understand why. This is correct behaviour, but important to be aware ",
                "of.",
            )

    # Normalize the data
    for t_tl in data_processed.keys():
        data_processed[t_tl]["value"] = (
            data_processed[t_tl]["value"] / data_processed[t_tl]["value"][0]
        )

    return data_processed


def isvalid_parameters_grid(parameters: Dict) -> bool:
    """Function checks whether the provided parameters are valid for GRID fitting.

    Parameters
    ----------
    parameters : Dict
        Dictionary containing all the parameters needed to perform the GRID fitting.

    Returns
    -------
    bool
        Defines whether the dictionary is valid or not.

    Notes
    -----
    Valid key, value pairs:

    "k": np.ndarray,
    "k_min": float,
    "k_max": float,
    "N": int,
    "reg_weight": float,
    "fit_a": bool,
    "a_fixed": None | float

    Example
    -------
    >>> parameters = {
            "k_min": 10**(-3),
            "k_max": 10**1,
            "N": 200,
            "reg_weight": 0.01
            "fit_a": True,
            "a_fixed": None
        }
    >>> isvalid_parameters_grid(parameters)
    True

    """
    KEYS_ALLOWED = set(
        [
            "k",
            "k_min",
            "k_max",
            "N",
            "reg_weight",
            "scale",
            "fit_a",
            "a_fixed",
        ]
    )

    param_keys = set(parameters.keys())

    if len(param_keys.difference(KEYS_ALLOWED)) != 0:
        raise ValueError(
            f"The keywords: {param_keys.difference(KEYS_ALLOWED)} are not valid \
                parameters t_tls"
        )

    elif "k_min" not in parameters:
        raise ValueError("Key 'k_min' is missing in the parameters variable")
    elif "k_max" not in parameters:
        raise ValueError("Key 'k_max' is missing in the parameters variable")
    elif "N" not in parameters:
        raise ValueError("Key 'N' is missing in the parameters variable")
    elif "scale" not in parameters:
        raise ValueError("Key 'scale' is missing in the parameters variable")
    elif "reg_weight" not in parameters:
        raise ValueError("Key 'reg_weight' is missing in the parameters variable")
    elif "fit_a" not in parameters:
        raise ValueError("Key 'fit_a' is missing in the parameters variable")

    if parameters["scale"] not in set(["log", "linear"]):
        raise ValueError("Value for key 'scale' should be either log or linear")
    if not parameters["fit_a"] and "a_fixed" not in parameters:
        raise ValueError(
            "If the photobleaching number should not be fitted then a photobleaching \
                number (a) should be provided"
        )
    return True


def isvalid_parameters_n_exp(parameters: Dict) -> bool:
    """Function checks whether the provided parameters are valid for n-exp fitting.

    Parameters
    ----------
    parameters : Dict
        Dictionary containing all the parameters needed to perform the n-exp fitting.

    Returns
    -------
    bool
        Defines whether the dictionary is valid or not.

    Notes
    -----
    Valid key, value pairs:
    "n_exp": int | Sequence
    "k": np.ndarray
    "k_min": float
    "k_max": float
    "fit_a": bool
    "a_fixed": None | float

    Example
    -------
    >>> parameters = {
            "n_exp": 3
            "k_min": 10**(-3),
            "k_max": 10**1,
            "fit_a": True,
            "a_fixed": None
        }
    >>> isvalid_parameters_n_exp(parameters)
    True
    """
    KEYS_ALLOWED = set(
        [
            "n_exp",
            "k",
            "k_min",
            "k_max",
            "fit_a",
            "a_fixed",
        ]
    )

    param_keys = set(parameters.keys())

    if len(param_keys.difference(KEYS_ALLOWED)) != 0:
        raise ValueError(
            f"The keywords: {param_keys.difference(KEYS_ALLOWED)} are not valid \
                parameters t_tls"
        )
    elif "n_exp" not in parameters:
        raise ValueError("Key 'n_exp' is missing in the parameters variable")
    elif "k" not in parameters and (
        "k_min" not in parameters or "k_max" not in parameters
    ):
        if "k_min" not in parameters:
            raise ValueError("Key 'k_min' is missing in the parameters variable")
        elif "k_max" not in parameters:
            raise ValueError("Key 'k_max' is missing in the parameters variable")
    elif "fit_a" not in parameters:
        raise ValueError("Key 'fit_a' is missing in the parameters variable")
    if not parameters["fit_a"] and "a_fixed" not in parameters:
        raise ValueError(
            "If the photobleaching number should not be fitted then a photobleaching \
                number (a) should be provided"
        )

    return True


def split_parameters(parameters: Dict) -> Dict:
    """Function splits a provided parameters dictionary into two different
    dictionaries. One dictionary for parameters required for GRID fitting and one
    dictionary for parameters required for n-exp fitting."""
    KEYS_GRID = set(
        [
            "k",
            "k_min",
            "k_max",
            "N",
            "reg_weight",
            "scale",
            "fit_a",
            "a_fixed",
        ]
    )

    KEYS_N_EXP = set(
        [
            "n_exp",
            "k",
            "k_min",
            "k_max",
            "fit_a",
            "a_fixed",
        ]
    )

    parameters_grid = dict()
    parameters_n_exp = dict()

    for key, value in parameters.items():

        if key in KEYS_GRID:
            parameters_grid[key] = value
        if key in KEYS_N_EXP:
            parameters_n_exp[key] = value

    return parameters_grid, parameters_n_exp



"""
Module with functions to compute curves or survival functions.
"""
import math
from typing import Dict, List, Tuple, Union

import numpy as np

from . import calc


def retrieve_track_lifes(data, filtered: bool = True) -> Dict:
    """Function retrieves the amount of frames a molecule is bound.

    Parameters
    ----------
    data : List[List]
        The data values as follows: (f, t, x, y, track_id, disp, intensity, sigma, fit_error)
    filtered : bool (CURRENTLY NOT IMPLEMENTED!!!)
        Only return residence frame lengths when the full tracklet is imaged. You don't know
        how long a particle has been bound at the start of imaging or how long it will stay bound
        after the imaging stops, so these particle residence frame length are not really
        trustworthy.

    Returns
    -------
    track_lifes: np.array
    """
    # TODO: filtering is currently not implemented
    track_lifes = []

    frame_start = None
    frame_end = None

    for i in range(len(data)):

        # if there is no value for the frame_start and frame_end, set a value, ie. first data point
        if frame_start is None:
            frame_start = data[i][0]
            frame_end = data[i][0]
        elif data[i][4] == data[i - 1][4]:
            frame_end = data[i][0]
        else:
            # If frame_end=5, and frame_start=2, then track_life should be 3
            # ie. count the amount of delta t's between the start and end
            track_life = frame_end - frame_start
            track_lifes.append(track_life)
            frame_start = data[i][0]
            frame_end = data[i][0]

        # if it is the last data point, add the track life
        if i == len(data) - 1:
            track_life = frame_end - frame_start
            track_lifes.append(track_life)

    return track_lifes


# TODO: update docstring
def compute_survival_function(
    track_lifes: Union[List, np.ndarray],
    t_tl: float,
    min_track_life: int = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Function creates the time and value arrays describing the survival function distribution from
    a set of track lifes.


    # If frame_end=5, and frame_start=2, then track_life should be 3
    # ie. count the amount of delta t's between the start and end

    So track lifes is the number of delta t's.

    Parameters
    ----------
    track_lifes : List or np.ndarray
        Contains all the individual track lifes, so duplicate track life values are
        allowed/expected.
    t_tl : float
        Time-lapse time in seconds.
    min_track_life : int
        The minimum track life length, this is inferred when the value is set to None, and then it
        is assumed that the minimum value in the track_lifes list/np.ndarray is equal to the minimum
        allowable value set in the settings. However, this can also be set manually. (default None)

    Returns
    -------
    time: np.ndarray
        The time vector indicating all the time points for the values of the survival function.
    sf_value: np.ndarray
        The survival function values for their respective time point.
    """
    if isinstance(track_lifes, list):
        track_lifes = np.array(track_lifes, dtype=np.int64)

    if not np.issubdtype(track_lifes.dtype, np.integer):
        raise ValueError("The dtype of the provided array should be integer.")

    if min_track_life is None:
        # Get the shortest possible track life from all the track lifes. min_track_length assumes
        # that the minimum allowable track length is in the dataset, so if the minimum track length
        # was set to 3 than we assume that a track life of 3 is at least in the dataset once,
        # otherwise the min_track_length should be set in the function call.
        min_track_life = np.amin(track_lifes)

    # Get the largest possible track in this tl condition
    max_track_life = np.amax(track_lifes)

    # Retrieve the unique track length values and their counts
    track_life_values, count = np.unique(track_lifes, return_counts=True)

    # Array to store the number of occurences of each lifetime.
    amount = np.zeros(max_track_life + 1, dtype=np.int64)

    # Fill the array at the indices of the the unique track lengths with their respective count
    amount[track_life_values] = count

    # Delete the first n rows to account for the min_track_length
    amount = amount[min_track_life:]

    # Create the time vector
    time = np.linspace(
        min_track_life * t_tl,
        max_track_life * t_tl,
        num=((max_track_life - min_track_life) + 1),
        dtype=np.float64,
    )

    # Calculate the survival function values
    value = np.cumsum(amount[::-1])[::-1]

    return time, value


def compute_multi_exp(
    k: np.ndarray, s: np.ndarray, a: float, time: np.ndarray
) -> np.ndarray:
    """Function computes the multi-exponential for the given time points.

    Parameters
    ----------
    k : np.ndarray
        Decay rates
    s : np.ndarray
        Amplitudes of the respective decay rates. The sum of the amplitudes should
        be equal to zero.
    a : float
        Photobleaching number (a * t_int)
    time : np.ndarray
        Time points for which the multi-exponential needs to be determined.

    Returns
    -------
    value : np.ndarray
        The resulting normalized multi-exponential values at the given time points.
    """

    if not math.isclose(1.0, np.sum(s)):
        raise ValueError(f"Sum of amplitudes should be equal to 1, but is {np.sum(s)}.")

    # Frame number, start counting at 1
    m = np.arange(time.shape[0]) + 1

    # Calculate the model function
    # Decay spectrum
    h = calc.calch(time, k, s)
    # Bleaching
    q = np.exp(-a * m)

    # Calculate the complete GRID fit function
    value = q / q[0] * h / h[0]

    return value


def compute_grid_curve(
    k: np.ndarray, s: np.ndarray, a: float, time: np.ndarray
) -> np.ndarray:
    """Function computes the GRID curve for the given time points. This is just a
    wrapper function for :func:`compute_multi_exp`, since the GRID curve is just a
    multi-exponential curve.

    Parameters
    ----------
    k : np.ndarray
        Decay rates
    s : np.ndarray
        Amplitudes of the respective decay rates. The sum of the amplitudes should
        be equal to zero.
    a : float
        Photobleaching number (a * t_int)
    time : np.ndarray
        Time points for which the multi-exponential needs to be determined.

    Returns
    -------
    value : np.ndarray
        The resulting normalized multi-exponential values at the given time points.
    """

    value = compute_multi_exp(k, s, a, time)

    return value


def compute_multi_exp_for_data(
    fit_values_multi_exp: Dict[str, Union[np.array, float]],
    data: Dict[str, Dict[str, np.ndarray]],
):
    """Function computes the multi-exponential curves given the required parameters.

    Parameters
    ----------
    fit_values_multi_exp: Dict[str, Union[np.array, float]]
        Dictionary with the following key-value pairs:
            "k": np.ndarray with the dissociation rates
            "s": np.ndarray with the corresponding weights
            "a": bleaching number (a = k_b * t_int)
            Optional:
                if GRID fit:
                "fit_error": error of the fit
                if multi-exponential fit:
                "Adj_R_squared": adjusted R squared for every k and S
    data: Dict[str, Dict[str, np.ndarray]]
        Survival function data for time-lapse conditions with the following data
        structure:
            {
                "t_tl": {
                    "time": np.ndarray with the time points,
                    "value": np.ndarray with the survival function values,
                },
                ...
            }

    Returns
    -------
    data_multi_exp: Dict[str, Dict[str, np.ndarray]]
        Survival function data computed with the provided multi-exponential fit values
        for every time-lapse conditions in data. The object has the following data
        structure:
            {
                "t_tl": {
                    "time": np.ndarray with the time points,
                    "value": np.ndarray with the survival function values,
                }
            }
    """
    k = fit_values_multi_exp["k"]
    s = fit_values_multi_exp["s"]
    a = fit_values_multi_exp["a"]

    # Dictionary to store the GRID fit results
    data_multi_exp = dict()

    for t_tl in data.keys():

        data_multi_exp[t_tl] = dict()

        # Time points
        t = data[t_tl]["time"]

        # Store the time points
        data_multi_exp[t_tl]["time"] = t

        data_multi_exp[t_tl]["value"] = compute_multi_exp(k, s, a, t)

    return data_multi_exp


def compute_grid_curves_for_data(
    fit_values_grid: Dict[str, Union[np.array, float]],
    data: Dict[str, Dict[str, np.ndarray]],
):
    """Function computes the GRID curves for the given fit_values_grid parameters.
    This is a convenience function and is just a wrapper around the function
    :func:`calc_multi_exp`.

    Parameters
    ----------
    fit_values_grid: Dict[str, Union[np.array, float]]
        Dictionary with the following key-value pairs:
            "k": np.ndarray with the dissociation rates
            "s": np.ndarray with the corresponding weights
            "a": bleaching number (a = k_b * t_int)
            Optional:
                "fit_error": error of the fit
    data: Dict[str, Dict[str, np.ndarray]]
        Survival function data for time-lapse conditions with the following data
        structure:
            {
                "t_tl": {
                    "time": np.ndarray with the time points,
                    "value": np.ndarray with the survival function values,
                },
                ...
            }

    Returns
    -------
    data_grid: Dict[str, Dict[str, np.ndarray]]
        Survival function data computed with the provided GRID fit values for every
        time-lapse conditions in data. The object has the following data structure:
            {
                "t_tl": {
                    "time": np.ndarray with the time points,
                    "value": np.ndarray with the survival function values,
                },
                ...
            }
    """

    data_grid = compute_multi_exp_for_data(fit_values_grid, data)

    return data_grid



"""
Module with functions to read and write tracking data
"""
import pathlib
from typing import Tuple, Union, List
import csv

# TODO: update this to make it more general


def read_track_file_csv(path: Union[str, pathlib.Path]) -> List[Tuple]:
    """Function reads track data from a .csv file that has the data format of track files created
    by the TrackIt framework and returns the data.

    Parameters
    ----------
    path: str or pathlib.Path
        Path to the file with track information.

    Returns
    -------
    data: list of tuples
        List with tuples, where every tuple contains the information of one part of a track. Every
        tuple has the following structure: (f, t, x, y, track_id, disp, intensity, sigma, fit_error).
            f (int): frame number
            t (float): time
            x (float): x-coordinate in um
            y (float): y-coordinate in um
            track_id (int): id of the track
            disp (float): displacement in um between current localization of track and previous one.
                The value is -1 if it is the first localization of the track.
            intensity (float): TODO is from SOS
            sigma (float): TODO is from SOS
            fit_error (float): TODO is from SOS

    Raises
    ------
    ValueError
        If the path suffix does not end with .csv
    FileNotFoundError
        If the file is not present at path.
    """

    if isinstance(path, str):
        path = pathlib.Path(path)

    if path.suffix != ".csv":
        raise ValueError(f"path should end with .csv, but ends with {path.suffix}")

    data = []

    with open(path, "r") as csvfile:
        reader = csv.DictReader(csvfile)
        for i, row in enumerate(reader):
            if len(row) == 6:
                f = int(float(row["frame"]))
                t = float(row["t"])
                x = float(row["x"])
                y = float(row["y"])
                track_id = int(float(row["trajectory"]))

                # If the track_id is the same as the previous point
                # then we can calculate a displacement otherwise set at -1
                if i != 0 and track_id == data[-1][3]:
                    disp = ((x - data[-1][1]) ** 2 + (y - data[-1][2]) ** 2) ** 0.5
                else:
                    disp = -1

                intensity = None
                sigma = None
                fit_error = None
                data.append((f, t, x, y, track_id, disp, intensity, sigma, fit_error))
            else:
                print(
                    f"Something went wrong at {path}, line {i} does not have 6 values."
                )

    return data


def write_track_file_csv(path: Union[str, pathlib.Path], data: List[Tuple]):
    """Function writes track data to a .csv file with the data format of track files created
    by the TrackIt framework.

    Parameters
    ----------
    path: str or pathlib.Path
        Path to the file with track information.

    data: list of tuples
        List with tuples, where every tuple contains the information of one part of a track. Every
        tuple has the following structure: (f, t, x, y, track_id, disp, intensity, sigma, fit_error).
            f (int): frame number
            t (float): time
            x (float): x-coordinate in um
            y (float): y-coordinate in um
            track_id (int): id of the track
            disp (float): displacement in um between current localization of track and previous one.
                The value is -1 if it is the first localization of the track.
            intensity (float): TODO is from SOS
            sigma (float): TODO is from SOS
            fit_error (float): TODO is from SOS

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If the path suffix does not end with .csv
    """

    if isinstance(path, str):
        path = pathlib.Path(path)

    if path.suffix != ".csv":
        raise ValueError(f"path should end with .csv, but ends with {path.suffix}")

    with open(path, "w", newline="") as csvfile:
        fieldnames = ["", "frame", "t", "trajectory", "x", "y"]
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()

        for i, d in enumerate(data):
            f, t, x, y, track_id, disp, intensity, sigma, fit_error = d
            writer.writerow(
                {"": i, "frame": f, "t": t, "trajectory": track_id, "x": x, "y": y}
            )
    print(f"Writing data to {path} is finished.")


def read_track_file_txt(path: Union[str, pathlib.Path]) -> List[Tuple]:
    """Function reads track data from a .txt file that has the data format of track files created
    by SOS and returns the data.

    Parameters
    ----------
    path: str or pathlib.Path
        Path to the file with track information.

    Returns
    -------
    data: list of tuples
        List with tuples, where every tuple contains the information of one part of a track. Every
        tuple has the following structure: (f, t, x, y, track_id, disp, intensity, sigma, fit_error).
            f (int): frame number
            t (float): time
            x (float): x-coordinate in pixel value
            y (float): y-coordinate in pixel value
            track_id (int): id of the track
            disp (float): displacement in pixel values between current localization of track and
                previous one. The value is -1 if it is the first localization of the track.
            intensity (float): TODO is from SOS
            sigma (float): TODO is from SOS
            fit_error (float): TODO is from SOS

    Raises
    ------
    ValueError
        If the path suffix does not end with .txt
    FileNotFoundError
        If the file is not present at path.
    """

    if isinstance(path, str):
        path = pathlib.Path(path)

    if path.suffix != ".txt":
        raise ValueError(f"path should end with .txt, but ends with {path.suffix}")

    data = []
    # data = [None] * sum(1 for line in open(path, "r"))
    i = 0
    with open(path, "r") as rf:
        for line in rf:
            line = line.strip()
            values = line.split("\t")
            if len(values) == 8:
                f = int(float(values[0]))
                t = None
                x = float(values[1])
                y = float(values[2])
                track_id = int(float(values[3]))
                disp = float(values[4])
                intensity = float(values[5])
                sigma = float(values[6])
                fit_error = float(values[7])
                data.append((f, t, x, y, track_id, disp, intensity, sigma, fit_error))
            else:
                print(
                    f"Something went wrong at {path}, line {i} does not have 8 values."
                )
            i += 1
    return data



from .survival_function import read_data_survival_function, write_data_survival_function
from .tracks import read_track_file_csv

__all__ = [
    "read_data_survival_function",
    "write_data_survival_function",
    "read_track_file_csv",
]



"""
Module with functions to read and write fit results.
"""
import pathlib
from typing import Dict, Tuple, Union, List

import numpy as np
import scipy.io as sio

# TODO: likely not going to do .csv, .mat files seem great kind of hdf5 file format
# TODO: add function to save fit results to .csv file (also as pickle file??)
# TODO: add function to read fit results to .csv file
# TODO: add function to read and write fit results from .mat file from original GRID
# package


def write_fit_results(
    path: Union[str, pathlib.Path],
    fit_results: Tuple[Dict[str, Union[np.ndarray, float]]],
):
    """Function writes fit results to a .mat file."""

    if isinstance(path, str):
        path = pathlib.Path(path)

    sio.savemat(path, fit_results)
    print(f"Fit results are saved in {path}")


def read_fit_results(
    path: Union[str, pathlib.Path]
) -> Dict[str, Dict[str, Union[np.ndarray, float]]]:
    """Function reads the fit results from a .mat file."""

    mat_contents = sio.loadmat(path, simplify_cells=True)

    return mat_contents


def read_fit_results_trackit(
    path: Union[str, pathlib.Path]
) -> Dict[str, Dict[str, Union[np.ndarray, float]]]:
    """Function loads and parses the fit results from the original GRID package,
    which saves the data in a different data structure, so hence there is some extra
    parsing involved."""
    # TODO: improve documentation

    if isinstance(path, str):
        path = pathlib.Path(path)

    mat_contents = sio.loadmat(path, simplify_cells=True)

    # Create the dictionary with the spectrum values
    fit_results_grid = dict()
    fit_results_grid["k"] = mat_contents["spectrum"]["dissociation_rates"]
    fit_results_grid["S"] = mat_contents["spectrum"]["Spectrum"]
    fit_results_grid["a1"] = mat_contents["spectrum"]["bleachingnumber_1"]
    fit_results_grid["error_test"] = mat_contents["spectrum"]["error_test"]

    # Single-exponential data
    fit_results_single_exp = dict()
    fit_results_single_exp["k"] = np.array(
        [mat_contents["monoexponential"]["dissociation_rate"]], dtype=np.float64
    )
    fit_results_single_exp["s"] = np.ones(1, dtype=np.float64)
    fit_results_single_exp["a"] = mat_contents["monoexponential"]["bleachingnumber_1"]
    fit_results_single_exp["adj_R_squared"] = mat_contents["monoexponential"][
        "Adj_R_Sqared"
    ]

    # Double-exponential data
    fit_results_double_exp = dict()
    fit_results_double_exp["k"] = mat_contents["twoexponential"]["dissociation_rates"]
    fit_results_double_exp["s"] = mat_contents["twoexponential"]["Amplitudes"]
    fit_results_double_exp["a"] = mat_contents["twoexponential"]["bleachingnumber_1"]
    fit_results_double_exp["adj_R_squared"] = mat_contents["twoexponential"][
        "Adj_R_Sqared"
    ]

    # Triple-exponential data
    fit_results_triple_exp = dict()
    fit_results_triple_exp["k"] = mat_contents["threeexponential"]["dissociation_rates"]
    fit_results_triple_exp["s"] = mat_contents["threeexponential"]["Amplitudes"]
    fit_results_triple_exp["a"] = mat_contents["threeexponential"]["bleachingnumber_1"]
    fit_results_triple_exp["adj_R_squared"] = mat_contents["threeexponential"][
        "Adj_R_Sqared"
    ]

    # Throw all the results into one dictionary
    fit_results = {
        "grid": fit_results_grid,
        "1-exp": fit_results_single_exp,
        "2-exp": fit_results_double_exp,
        "3-exp": fit_results_triple_exp,
    }

    return fit_results


def read_data_grid_resampling(
    path: str,
) -> Tuple[
    Dict[str, Union[np.ndarray, float]], List[Dict[str, Union[np.ndarray, float]]]
]:
    """Function loads and parses the resampling data from GRID.

    Parameters:
    ----------
    path: str
        Path to the file with all the resampling data

    Returns:
    -------
    results_full: Dict[str, Union[np.array, float]]
        Dictionary with the following key-value pairs:
            "k": np.ndarray with the dissociation rates
            "S": np.ndarray with the corresponding weights
            "a": bleaching number (a = kb * t_int)
            "loss": final cost value
    results_resampling: List[Dict[str, Union[np.array, float]]]
        List of dictionaries with the following key-value pairs:
            "k": np.ndarray with the dissociation rates
            "s": np.ndarray with the corresponding weights
            "a": bleaching number (a = kb * t_int)
            "loss": final cost value
        Every dictionary entry in the list contains the results of one data resample.
    """
    mat_contents = sio.loadmat(path, simplify_cells=True)

    results_full = mat_contents["result100"]
    results_resampling = mat_contents["results"]
    return results_full, results_resampling


def read_data_grid_resampling_trackit(
    path: str,
) -> Tuple[
    Dict[str, Union[np.ndarray, float]], List[Dict[str, Union[np.ndarray, float]]]
]:
    """Function loads and parses the resampling data from GRID.

    Parameters:
    ----------
    path: str
        Path to the file with all the resampling data

    Returns:
    -------
    results_full: Dict[str, Union[np.array, float]]
        Dictionary with the following key-value pairs:
            "k": np.ndarray with the dissociation rates
            "S": np.ndarray with the corresponding weights
            "a1": bleaching number (a = kb * t_int)
    results_resampling: List[Dict[str, Union[np.array, float]]]
        List of dictionaries with the following key-value pairs:
            "k": np.ndarray with the dissociation rates
            "S": np.ndarray with the corresponding weights
            "a1": bleaching number (a = kb * t_int)
        Every dictionary entry in the list contains the results of one data resample.
    """
    mat_contents = sio.loadmat(path, simplify_cells=True)

    results_full = mat_contents["result100"]
    results_resampling = mat_contents["results"]
    return results_full, results_resampling



"""
Module with functions to plot event spectrum and state spectrum.
"""
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.offsetbox import AnchoredText

from . import _plot_utils


def _base_spectrum(
    key_to_k_and_weight,
    scale: str = "log",
    threshold: float = 0.0,
    xlim: Tuple[float, float] = None,
    ylim: Tuple[float, float] = None,
    figsize: Tuple[float, float] = (8, 5),
    color=None,
    add_legend: bool = True,
):
    """Base function that plots the spectra in one figure.

    Parameters
    ----------
    key_to_k_and_weight : Dict[str, Dict[str, np.ndarray]]
        The decay rates and weights wrapped in the following structure:
            {
                "label": {
                    "k": np.ndarray with the decay rates,
                    "weight": np.ndarray with the respective weights
                }
            }
    scale : {"log", "linear"}, optional
        The scale on the x-axis. The default value is "log".
    threshold : float, optional
        The weight threshold for plotting lines, by default 0.0.
    xlim : Tuple[float, float], optional
        A tuple of the x-axis limits, by default None.
    ylim : Tuple[float, float], optional
        A tuple of the y-axis limits, by default None.
    figsize : Tuple[float, float], optional
        A tuple of the figure size, by default (8, 5).
    color : color or sequence of colors, optional
        The color or colors to use for the plotting. The standard gridlib colors are
        used if the value is set to None. The value is by default None.
    add_legend : bool, optional
        Indicates whether the legend needs to be plotted, by default True.

    Returns
    -------
    fig : matplotlib.Figure
        TODO: link this to matplotlib.Figure documentation
        https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure
    ax: matplotlib.axes.Axes
        TODO: link to the matplotlib.axes.Axes documentation
        https://matplotlib.org/stable/api/axes_api.html#matplotlib.axes.Axes
    """

    # Check if the scale option is valid
    if scale not in ["log", "linear"]:
        raise ValueError("The variable 'scale' does not contain a valid value.")

    # Colors
    gridlib_colors = _plot_utils._gridlib_colors()
    if color is None and len(key_to_k_and_weight) <= len(gridlib_colors):
        color = gridlib_colors[:]
    elif color is None:
        color = _plot_utils._default_colors()

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)

    # Retrieve the keys and sort them
    keys = list(key_to_k_and_weight.keys())
    keys.sort()
    if "grid" in keys:
        # First remove it
        keys.remove("grid")
        # and place it at the start of the list so it is plotted first
        keys.insert(0, "grid")

    for i, key in enumerate(keys):
        k = key_to_k_and_weight[key]["k"]
        weight = key_to_k_and_weight[key]["weight"]

        if key == "grid":
            linestyle_vlines = "solid"
            label = "GRID spectrum"
        else:
            linestyle_vlines = "dashed"
            label = key

        idx = weight >= threshold
        ax.plot(
            k[idx],
            weight[idx],
            linestyle="None",
            marker="o",
            markersize=3,
            color=color[i],
            label=label,
        )
        ax.vlines(
            k[idx],
            np.zeros(k.shape[0])[idx],
            weight[idx],
            linewidth=1,
            linestyles=linestyle_vlines,
            color=color[i],
            label=label,
        )

    if scale == "log":
        ax.set_xscale("log")

    # Axis limits
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    if add_legend:
        # Legend
        # Remove duplicate labels
        # Implementation from: https://stackoverflow.com/questions/13588920/stop-matplotlib-repeating-labels-in-legend
        handles, labels = ax.get_legend_handles_labels()
        by_label = dict(zip(labels, handles))
        ax.legend(by_label.values(), by_label.keys())  # , loc="lower left")

    return fig, ax


# TODO: think about setting the weight for a single-exponential to 0.2 for visual reasons


def event_spectrum(
    fit_values,
    scale: str = "log",
    threshold: float = 0.0,
    xlim: Tuple[float, float] = None,
    ylim: Tuple[float, float] = None,
    figsize: Tuple[float, float] = (8, 5),
    color=None,
    add_legend: bool = True,
):
    """Function plots the event spectrum of different fit values in one figure.

    Parameters
    ----------
    key_to_k_and_weight : Dict[str, Dict[str, np.ndarray]]
        The decay rates and weights wrapped in the following structure:
            {
                "label": {
                    "k": np.ndarray with the decay rates,
                    "weight": np.ndarray with the respective weights
                }
            }
    scale : {"log", "linear"}, optional
        The scale on the x-axis. The default value is "log".
    threshold : float, optional
        The weight threshold for plotting lines, by default 0.0.
    xlim : Tuple[float, float], optional
        A tuple of the x-axis limits, by default None.
    ylim : Tuple[float, float], optional
        A tuple of the y-axis limits, by default None.
    figsize : Tuple[float, float], optional
        A tuple of the figure size, by default (8, 5).
    color : color or sequence of colors, optional
        The color or colors to use for the plotting. The standard gridlib colors are
        used if the value is set to None. The value is by default None.
    add_legend : bool, optional
        Indicates whether the legend needs to be plotted, by default True.

    Returns
    -------
    fig : matplotlib.Figure
        TODO: link this to matplotlib.Figure documentation
        https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure
    ax: matplotlib.axes.Axes
        TODO: link to the matplotlib.axes.Axes documentation
        https://matplotlib.org/stable/api/axes_api.html#matplotlib.axes.Axes
    """

    a = None  # Photobleaching number

    key_to_k_and_weight = dict()
    for key in fit_values.keys():

        k = fit_values[key]["k"]
        s = fit_values[key]["s"]

        key_to_k_and_weight[key] = dict()
        key_to_k_and_weight[key]["k"] = k
        key_to_k_and_weight[key]["weight"] = s

        # Store the photobleaching number if it is in the fit results
        if key == "grid":
            a = fit_values[key]["a"]

    fig, ax = _base_spectrum(
        key_to_k_and_weight,
        scale=scale,
        threshold=threshold,
        xlim=xlim,
        ylim=ylim,
        figsize=figsize,
        color=color,
        add_legend=add_legend,
    )

    # Labels
    ax.set_xlabel("dissociation rate (1/s)")
    ax.set_ylabel("event spectrum")

    if a is not None:
        # add the bleaching number in the plot if there were grid fit values provided
        # https://stackoverflow.com/questions/23112903/matplotlib-text-boxes-automatic-position
        anchored_text = AnchoredText(f"a = {a:.5f}", loc="center left", frameon=False)
        ax.add_artist(anchored_text)

    return fig, ax


def state_spectrum(
    fit_values,
    scale: str = "log",
    threshold: float = 0.0,
    xlim: Tuple[float, float] = None,
    ylim: Tuple[float, float] = None,
    figsize: Tuple[float, float] = (8, 5),
    color=None,
    add_legend: bool = True,
):
    """Function plots the state spectrum of different fit values in one figure.

    Parameters
    ----------
    key_to_k_and_weight : Dict[str, Dict[str, np.ndarray]]
        The decay rates and weights wrapped in the following structure:
            {
                "label": {
                    "k": np.ndarray with the decay rates,
                    "weight": np.ndarray with the respective weights
                }
            }
    scale : {"log", "linear"}, optional
        The scale on the x-axis. The default value is "log".
    threshold : float, optional
        The weight threshold for plotting lines, by default 0.0.
    xlim : Tuple[float, float], optional
        A tuple of the x-axis limits, by default None.
    ylim : Tuple[float, float], optional
        A tuple of the y-axis limits, by default None.
    figsize : Tuple[float, float], optional
        A tuple of the figure size, by default (8, 5).
    color : color or sequence of colors, optional
        The color or colors to use for the plotting. The standard gridlib colors are
        used if the value is set to None. The value is by default None.
    add_legend : bool, optional
        Indicates whether the legend needs to be plotted, by default True.

    Returns
    -------
    fig : matplotlib.Figure
        TODO: link this to matplotlib.Figure documentation
        https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure
    ax: matplotlib.axes.Axes
        TODO: link to the matplotlib.axes.Axes documentation
        https://matplotlib.org/stable/api/axes_api.html#matplotlib.axes.Axes
    """

    a = None  # Photobleaching number

    key_to_k_and_weight = dict()
    for key in fit_values.keys():

        k = fit_values[key]["k"]
        s = fit_values[key]["s"]

        s_state = (1 / k) * s
        s_state = s_state / np.sum(s_state)  # normalization

        key_to_k_and_weight[key] = dict()
        key_to_k_and_weight[key]["k"] = k
        key_to_k_and_weight[key]["weight"] = s_state

        # Store the photobleaching number if it is in the fit results
        if key == "grid":
            a = fit_values[key]["a"]

    fig, ax = _base_spectrum(
        key_to_k_and_weight,
        scale=scale,
        threshold=threshold,
        xlim=xlim,
        ylim=ylim,
        figsize=figsize,
        color=color,
        add_legend=add_legend,
    )

    # Labels
    ax.set_xlabel("dissociation rate (1/s)")
    ax.set_ylabel("state spectrum")

    if a is not None:
        # add the bleaching number in the plot if there were grid fit values provided
        # https://stackoverflow.com/questions/23112903/matplotlib-text-boxes-automatic-position
        anchored_text = AnchoredText(f"a = {a:.5f}", loc="center right", frameon=False)
        ax.add_artist(anchored_text)

    return fig, ax



"""
Module with all the functions required for the GRID and multi-exponential
calculations.
"""
from typing import Dict, Tuple, Union

import numpy as np
from . import data_utils

# TODO: Make lsq_obj_multi_exp() (with n being the number of exponents in the wrapper function)


def transmat(time: np.ndarray, k: np.ndarray) -> np.ndarray:
    """Function to calculate the forward Laplace transformation matrix A (MxN)-
    matrix.

    Parameters
    ----------
    time: np.ndarray
        time points (m * t_tl)
    k: np.ndarray
        decay rates

    Returns
    -------
    A: np.ndarray
        Forward Laplace transformation matrix A (MxN) matrix
    """

    time = np.reshape(time, (time.shape[0], 1))
    k = np.reshape(k, (1, k.shape[0]))

    A = np.exp(-(time @ k))

    # m = time.shape[0]
    # n = k.shape[0]
    # A = np.ones((m, n))
    #
    # for i in range(m):
    #     for j in range(n):
    #         A[i, j] = np.exp(-k[j]*time[i])

    return A


def calch(time: np.ndarray, k: np.ndarray, s: np.ndarray) -> np.ndarray:
    """Function calculates the superposition of the different independent decay
    curves with the forward Laplace transformation.

    Parameters
    ----------
    time: np.ndarray
        The time points to calculate the forward Laplace transform for.
    k: np.ndarray
        Decay rates.
    s: np.ndarray
        Amplitudes of the decay rates.

    Returns
    -------
    h: np.ndarray
        Superposition of all the independent decay processes for the time points time.
    """

    # Transformation matrix
    A = transmat(time, k)

    # Weighted transformation
    h = A @ s

    return h


def gradh(
    eq0: np.ndarray, q: np.ndarray, h: np.ndarray, time: np.ndarray, k: np.ndarray
) -> np.ndarray:
    """Function calculates the gradient for the amplitudes of the decay rates.

    Parameters
    ----------
    eq0: np.ndarray
        Array with the distances between the model and the data.
    q: np.ndarray
        Array with values resulting from photobleaching.
    h: np.ndarray
        Array with values resulting from independent decay processes.
    time: np.ndarray
        Array with all the time points for which the values need to be calculated.
    k: np.ndarray
        Array with all the decay rates.
    Returns
    -------
    grad_s: np.ndarray
        Values for the gradient of the amplitudes for the respective decay rates in k.
    """

    # Derivatie with respect to Si
    A = transmat(time, k)
    A = A.T

    grad_s = np.zeros(k.shape[0])

    # Derivative
    for m in range(time.shape[0]):
        grad_s = grad_s + 2 * eq0[m] * q[m] / q[0] * (
            A[:, m] / h[0] - A[:, 0] * h[m] / (h[0]) ** 2
        )

    return grad_s


def gradq(
    eq0: np.ndarray, q: np.ndarray, h: np.ndarray, time: np.ndarray, a: float
) -> float:
    """Function calculates the gradient for the photobleaching.

    Parameters
    ----------
    eq0: np.ndarray
        Array with the distances between the model and the data.
    q: np.ndarray
        Array with values resulting from photobleaching.
    h: np.ndarray
        Array with values resulting from independent decay processes.
    time: np.ndarray
        Array with all the time points for which the values need to be calculated.
    a: np.ndarray
        Photobleaching number (k_b * t_int)

    Returns
    -------
    grad_a: float
        Gradient value for the photobleaching number.
    """

    grad_a = 0

    for m in range(eq0.shape[0]):
        grad_a = grad_a + 2 * eq0[m] * (h[m] / h[0]) * -m * np.exp(-a * m)
    return grad_a


def lsqobj_grid(
    values: np.ndarray,
    data: Dict[str, Dict[str, np.ndarray]],
    k: np.ndarray,
    reg_weight: float,
) -> Tuple[float, np.ndarray]:
    """Returns the loss value and the gradient for the GRID fit.

    Objective function for determining dissociation rate spectra. The cost function d
    contains the difference between the GRID fit and measured values and the
    regularization for the mean decay rate of the population.

    Parameters
    ----------
    values : np.ndarray
        Array with decay rate amplitudes + photobleaching number with the structure:
            values = np.array([s, a]) where s is an array with the amplitudes.
    data : Dict[str, Dict[str, np.ndarray]]
        Data of the survival function from real data with the following data structure:
        {
            "t_tl": {
                "time": np.ndarray with the time points,
                "value": np.ndarray with the survival function values,
            }
        }
    k : np.ndarray
        Array of the decay rates of the fixed grid positions
    reg_weight : float
        Regularization weight

    Returns
    -------
    d : float
        Loss value for the provided values.
    grad: np.ndarray
        Gradient for the provided values.
    """

    # Initialization of the variables
    # ---------------------------------------------------------------------------------
    d = 0
    ceq = 0  # Constrain equality
    gradceq = np.zeros(values.shape[0], dtype=np.float64)
    gradreg = np.zeros(values.shape[0], dtype=np.float64)

    # Unpack the values array and assign them to the correct variable
    # ---------------------------------------------------------------------------------
    s = values[: k.shape[0]]
    a = values[k.shape[0]]

    # Constraints for global fit
    # ---------------------------------------------------------------------------------

    # Cost function for difference between fit and measurement
    # Loop over all the time-lapse conditions

    tau_min = np.inf  # variable to store the minimum integration time
    for t_tl in data.keys():

        # Check if this is the minimum integration time (tau)
        tau = data_utils.get_time_sec(t_tl)
        if tau < tau_min:
            tau_min = tau

        # read n-th measured time-lapse
        time = data[t_tl]["time"]  # time-points (lifetime) array
        p = data[t_tl]["value"]  # probability array

        # IMPORTANT: frame count should start at 1, NOT 0
        m = np.arange(time.shape[0]) + 1

        # Calculate the model function
        # Decay spectrum
        h = calch(time, k, s)
        # Photobleaching
        q = np.exp(-a * m)

        # Calculate the squared differences
        eq0 = ((q / q[0]) * (h / h[0])) - (p / p[0])
        ceqt = eq0**2

        # Generate gradient
        gradhl = gradh(eq0, q, h, time, k)
        gradql = gradq(eq0, q, h, time, a)
        gradceqt = np.concatenate((gradhl, np.array([gradql])))

        # Bind constraint and gradient
        ceq = ceq + np.sum(ceqt)
        gradceq = gradceq + gradceqt

    # Regularization for the dead time
    # ------------------------------------------------------------------
    # Initialize variables
    tau = np.array([tau_min, 2 * tau_min], dtype=np.float64)  # death time
    A = transmat(tau, k)

    # Calculate the mean decay rate at borders t
    kquer = (A @ (k * s)) / (A @ s)

    # Regularisation
    reg = 0.5 * (kquer[0] - kquer[1]) ** 2

    # Gradient of regularisation
    temp0 = A[0, :].T / (A[0, :] @ s) * (k - kquer[0])
    temp1 = A[1, :].T / (A[1, :] @ s) * (k - kquer[1])
    gradreg[0 : s.shape[0]] = (kquer[0] - kquer[1]) * (temp0 - temp1)

    # Calculate the complete cost function
    # -------------------------------------------------------------------
    d = ceq + reg_weight * reg
    grad = gradceq + reg_weight * gradreg

    return d, grad


def global_multi_exp(
    values: np.ndarray,
    data: Dict[str, Dict[str, np.ndarray]],
    n: int,
) -> Tuple[float, np.ndarray]:
    """Function returns the residuals"""
    # Initialization of the variables
    # ---------------------------------------------------------------------------------
    d = np.array([], dtype=np.float64)

    # Unpack the values array and assign them to the correct variable
    # ---------------------------------------------------------------------------------
    k = values[:n]
    s = values[n : (2 * n)]
    a = values[-1]

    # Constraints for global fit
    # ---------------------------------------------------------------------------------

    # Cost function for difference between fit and measurement
    # Loop over all the time-lapse conditions
    for t_tl in data.keys():
        # read n-th measured time-lapse
        time = data[t_tl]["time"]  # time-points (lifetime) array
        p = data[t_tl]["value"]  # probability array

        # IMPORTANT: frame count should start at 1, NOT 0
        m = np.arange(time.shape[0]) + 1

        # Calculate the model function
        # Decay spectrum
        h = calch(time, k, s)
        # Photobleaching
        q = np.exp(-a * m)

        # Calculate the residuals
        eq0 = ((q / q[0]) * (h / h[0])) - (p / p[0])

        d = np.concatenate((d, eq0))

    return d


# TODO: make a multi-exp fitting with fixed k values function
def global_multi_exp_fixed_k(
    values: np.ndarray,
    k: np.ndarray,
    data: Dict[str, Dict[str, np.ndarray]],
    n: int,
) -> Tuple[float, np.ndarray]:
    """Function returns the residuals"""
    # Initialization of the variables
    # ---------------------------------------------------------------------------------
    d = np.array([], dtype=np.float64)

    # Unpack the values array and assign them to the correct variable
    # ---------------------------------------------------------------------------------
    s = values[:n]
    a = values[-1]

    # Constraints for global fit
    # ---------------------------------------------------------------------------------

    # Cost function for difference between fit and measurement
    # Loop over all the time-lapse conditions
    for t_tl in data.keys():
        # read n-th measured time-lapse
        time = data[t_tl]["time"]  # time-points (lifetime) array
        p = data[t_tl]["value"]  # probability array

        # IMPORTANT: frame count should start at 1, NOT 0
        m = np.arange(time.shape[0]) + 1

        # Calculate the model function
        # Decay spectrum
        h = calch(time, k, s)
        # Photobleaching
        q = np.exp(-a * m)

        # Calculate the residuals
        eq0 = ((q / q[0]) * (h / h[0])) - (p / p[0])

        d = np.concatenate((d, eq0))

    return d



# from .spectrum import event_spectrum, state_spectrum
from .spectrum import *
from .survival_function import data_multiple, data_vs_grid, data_vs_multi_exp

# __all__ = ["event_spectrum", "state_spectrum", ""]



"""Top-level package for GRIDLib."""

__author__ = """Boyd Christiaan Peters"""
__email__ = "boyd.c.peters@gmail.com"
__version__ = "0.3.4"

from . import io, plot
from .compute import *
from .data_utils import *
from .fit import *
from .resample import *
from .simulate import *



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gridlib', 'gridlib.io', 'gridlib.io.tests', 'gridlib.plot']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.4.0,<4.0.0', 'scipy>=1.6.0,<2.0.0']

setup_kwargs = {
    'name': 'gridlib',
    'version': '0.3.4',
    'description': 'Python package to perform GRID analysis on fluorescence survival time distributions.',
    'long_description': '\n# GRIDLib\n\n<div align="center">\n\n[![PyPI - Version](https://img.shields.io/pypi/v/gridlib.svg)](https://pypi.python.org/pypi/gridlib)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/gridlib.svg)](https://pypi.python.org/pypi/gridlib)\n[![Tests](https://github.com/boydcpeters/gridlib/workflows/tests/badge.svg)](https://github.com/boydcpeters/gridlib/actions?workflow=tests)\n[![Codecov](https://codecov.io/gh/boydcpeters/gridlib/branch/main/graph/badge.svg)](https://codecov.io/gh/boydcpeters/gridlib)\n[![Read the Docs](https://readthedocs.org/projects/gridlib/badge/)](https://gridlib.readthedocs.io/)\n[![PyPI - License](https://img.shields.io/pypi/l/gridlib.svg)](https://pypi.python.org/pypi/gridlib)\n\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n[![Contributor Covenant](https://img.shields.io/badge/Contributor%20Covenant-2.0-4baaaa.svg)](https://www.contributor-covenant.org/version/2/0/code_of_conduct/)\n\n</div>\n\nPython package to perform GRID analysis on fluorescence survival time distributions.\n\n* GitHub repo: <https://github.com/boydcpeters/gridlib.git>\n* Documentation: <https://gridlib.readthedocs.io>\n* Free software: GNU General Public License v3\n\n## Features\n\n* TODO\n\n## Quickstart\n\nTODO\n\n## References\n\nThe GRID fitting procedure implemented in this package is based on the following paper:\n\n```latex\n@article{reisser2020inferring,\n  title={Inferring quantity and qualities of superimposed reaction rates from single molecule survival time distributions},\n  author={Reisser, Matthias and Hettich, Johannes and Kuhn, Timo and Popp, Achim P and Gro{\\ss}e-Berkenbusch, Andreas and Gebhardt, J Christof M},\n  journal={Scientific reports},\n  volume={10},\n  number={1},\n  pages={1--13},\n  year={2020},\n  publisher={Nature Publishing Group}\n}\n```\n\n## Credits\n\nThis package was created with [Cookiecutter][cookiecutter] and the [fedejaure/cookiecutter-modern-pypackage][cookiecutter-modern-pypackage] project template.\n\n[cookiecutter]: https://github.com/cookiecutter/cookiecutter\n[cookiecutter-modern-pypackage]: https://github.com/fedejaure/cookiecutter-modern-pypackage\n',
    'author': 'Boyd Christiaan Peters',
    'author_email': 'boyd.c.peters@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/boydcpeters/gridlib',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



import csv
import pathlib
from typing import Dict, Union

import numpy as np

from .. import data_utils


def write_data_survival_function(
    path: Union[str, pathlib.Path], data: Dict[str, Dict[str, np.ndarray]]
):
    """Function writes the survival function data to a csv file.

    Parameters
    ----------
    path: str or pathlib.Path
        Path to the location of the file, where to save to. If the file does not exist,
        it will be created.
    data: Dict[str, Dict[str, np.ndarray]]
        Data of the survival function to write to a csv file. The dictionary structure
        is as follows:
        {
            "t_tl": {
                "time": np.ndarray with the time points,
                "value": np.ndarray with the survival function values,
            }
        }

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If the path suffix does not end with .csv

    Notes
    -----
    The .csv file with the survival time distributions will looks as follows
    (without the header):

    time-lapse 1, value, time-lapse 2, value, ...\n
    Delta t_1,f1(Delta t_1), Delta t_2, f2(Delta t_2)\n
    2*Delta t_1,f1(2*Delta t_1), 2*Delta t_2, f2(2*Delta t_2)\n
    3*Delta t_1,f1(3*Delta t_1), 3*Delta t_2, f2(3*Delta t_2)\n
    ..., ..., ..., ...
    """
    if isinstance(path, str):
        path = pathlib.Path(path)

    if path.suffix != ".csv":
        raise ValueError(f"path should end with .csv, but ends with {path.suffix}")

    # Make sure the units for all the time-lapse conditions is in seconds
    # This is needed for sorting of the values
    data = data_utils.fmt_t_str_data(data)

    t_tl_all = list(data.keys())

    # Sort the tl values from low to high.
    t_tl_all.sort()

    max_length = 0

    for t_tl in t_tl_all:
        val = data[t_tl]["time"].shape[0]
        if val > max_length:
            max_length = val

    rows_to_write = []
    for i in range(max_length):
        row = []
        for t_tl in t_tl_all:
            row.append(data[t_tl]["time"][i] if i < data[t_tl]["time"].shape[0] else "")
            row.append(
                data[t_tl]["value"][i] if i < data[t_tl]["value"].shape[0] else ""
            )
        row = tuple(row)
        rows_to_write.append(row)

    with open(path, "w", newline="") as csvfile:
        writer = csv.writer(csvfile)
        for row in rows_to_write:
            writer.writerow(row)
    print(f"Writing data to {path} is finished.")


def read_data_survival_function(
    path: Union[str, pathlib.Path]
) -> Dict[str, Dict[str, np.ndarray]]:
    """Function reads survival function from csv file.

    Parameters
    ----------
    path: str or pathlib.Path
        Path to the location of the .csv file containing the survival time
        distributions.

    Returns
    -------
    data: Dict[str, Dict[str, np.ndarray]]
        Data of the survival function to write to a csv file. The dictionary structure
        is as follows:
        {
            "t_tl": {
                "time": np.ndarray with the time points,
                "value": np.ndarray with the survival function values,
            }
        }

    Raises
    ------
    ValueError
        If the path suffix does not end with .csv
    FileNotFoundError
        If the file is not present at path.

    Notes
    -----
    The .csv file with the surivival time distributions should look like this
    (without the header):

    time-lapse 1, value, time-lapse 2, value, ...\n
    Delta t_1,f1(Delta t_1), Delta t_2, f2(Delta t_2)\n
    2*Delta t_1,f1(2*Delta t_1), 2*Delta t_2, f2(2*Delta t_2)\n
    3*Delta t_1,f1(3*Delta t_1), 3*Delta t_2, f2(3*Delta t_2)\n
    ..., ..., ..., ...
    """
    if isinstance(path, str):
        path = pathlib.Path(path)

    if path.suffix != ".csv":
        raise ValueError(f"path should end with .csv, but ends with {path.suffix}")

    data = dict()

    with open(path, "r", newline="") as csvfile:
        reader = csv.reader(csvfile)

        # Set-up the data dictionary
        row1 = next(reader)  # read the first row
        row2 = next(reader)  # read the second row
        # determine the number of tl conditions, every time-lapse condition consist of
        # two columns, the time column and the value column
        num_tls = len(row1) // 2

        for i in range(num_tls):
            # Time-lapse condition is determined by calculation the difference
            # between the first and second time point.
            time_1 = float(row1[2 * i])
            time_2 = float(row2[2 * i])

            # Store the initial key
            time_s = round(time_2 - time_1, 5)
            t_tl = f"{time_s}s"
            data[t_tl] = dict()

        # Format the keys correctly, it is possible that the key due to rounding
        # was 1.00000s, so now format them to 1s
        data = data_utils.fmt_t_str_data(data)

        # Set the pointer back to the start of the file
        csvfile.seek(0)

        for row in reader:
            for i, t_tl in zip(range(num_tls), data.keys()):

                # Read-out the time value
                time_s_str = row[(2 * i)]
                if time_s_str != "":
                    temp = data[t_tl].get("time", [])
                    time_s = round(float(time_s_str), 4)
                    temp.append(time_s)
                    data[t_tl]["time"] = temp

                # Read out the distribution value
                val_str = row[(2 * i) + 1]
                if val_str != "":
                    temp = data[t_tl].get("value", [])
                    val = int(val_str)
                    temp.append(val)
                    data[t_tl]["value"] = temp

        # After the lists are completely filled, loop over all the lists and convert
        # them to arrays
        for t_tl in data.keys():
            for key in data[t_tl].keys():
                data[t_tl][key] = np.array(data[t_tl][key])

    return data



"""
Module with function to sample from a provided GRID spectrum or multi-exponential.
"""
from typing import Sequence, Union

import numpy as np

from . import data_utils


def tl_simulation_single(
    k: np.ndarray,
    s: np.ndarray,
    kb: float,
    t_int: float,
    t_tl: float,
    N: int = 10000,
):
    """Function simulates the dissociation of molecules and calculates the resulting
    survival time distribution.

    Parameters
    ----------
    k: np.ndarray
        Decay rates with units per second.
    s: np.ndarray
        Amplitudes for the respective decay rates.
    kb: float
        Photobleaching rate per second (kb = a / t_int).
    t_int: float
        The integration time in seconds.
    t_tl: float
        The time-lapse time to simulate in seconds.
    N: int, optional
        Number of molecules to simulate and use for the survival time distribution
        calculation (default 10000).

    Returns
    -------
    Dict[str, Dict[str, np.ndarray]]
        Survival function data for specified time-lapse condition with the following
        data structure:
            {
                "t_tl": {
                    "time": np.ndarray with the time points,
                    "value": np.ndarray with the survival function values,
                }
            }
    """

    p = s / np.sum(s)

    # Can also change kb * t_int to a
    keff = (kb * t_int) / t_tl + k

    count = 0
    time = np.linspace(t_tl, 1000 * t_tl, num=1001, endpoint=True)

    binding = np.zeros(time.shape[0])
    binding_sum = 0

    # TODO: future - improve speed: multiprocessing, or use "chunks", so do like
    # 1000 molecules at the same time with arrays?
    rng = np.random.default_rng()

    # Simulate until enough points have been captured or the computation time is
    # exceeded
    while binding_sum < N and count < (N * 10):

        keff_state = rng.choice(keff, p=p)
        lifetime = rng.exponential(scale=1 / keff_state)

        if lifetime > time[0]:
            idx = np.sum(lifetime > time) - 1  # - 1 because indexing starts at 0
            binding[idx] = binding[idx] + 1
            binding_sum = binding_sum + 1
        else:
            count = count + 1

    # Calculate the survival function values with cumulative summing
    value = np.cumsum(binding[::-1])[::-1]

    # Delete the empty elements
    time = time[value > 0]
    value = value[value > 0]

    data = {f"{t_tl}s": {"time": time, "value": value}}

    # Make sure the data is formatted correctly
    data = data_utils.fmt_t_str_data(data)

    return data


def tl_simulation(
    k: np.ndarray,
    s: np.ndarray,
    kb: float,
    t_int: float,
    t_tl_all: Sequence[Union[float, str]],
    N: Union[int, Sequence[int]] = 10000,
):
    """Function simulates the dissociation of molecules and calculates the resulting
    survival time distribution.

    Parameters
    ----------
    k : np.ndarray
        Decay rates with units per second.
    s : np.ndarray
        Amplitudes for the respective decay rates.
    kb : float
        Photobleaching rate per second (kb = a / t_int).
    t_int : float
        The integration time in seconds.
    t_tl_all : Sequence[float | str]
        The time-lapse times to simulate. If the time-lapse times are provided as
        floats then they are interpreted as seconds. However, the time-lapse times can
        also be provided as strings, eg. '100ms' or '1.5s'.
    N : int | Sequence[int], optional
        Number of molecules to simulate and use for the survival time distribution
        calculation (default 10000). If only an integer value is provided then this
        value will be used for all the simulations. In the case that the provided
        N object is an iterable, each element is used as value for each time-lapse
        condition.

    Returns
    -------
    Dict[str, Dict[str, np.ndarray]]
        Survival function data for specified time-lapse conditions with the following
        data structure:
            {
                "t_tl": {
                    "time": np.ndarray with the time points,
                    "value": np.ndarray with the survival function values,
                },
                ...
            }
    """
    # Dictionary with all the simulated data points
    data = dict()

    if isinstance(N, Sequence):
        if len(N) != len(t_tl_all):
            raise ValueError(
                f"Number of `N` values provided ({len(N)}) is not equal to the number "
                f"of provided time-lapse times provided in `t_tl_all` ({len(t_tl_all)})."
            )
    else:
        N = [N for _ in range(len(t_tl_all))]

    # TODO: future - allow this to be done in a multiprocessed way
    for i, t_tl in enumerate(t_tl_all):
        if isinstance(t_tl, str):
            t_tl_s = data_utils.get_time_sec(t_tl)
        elif isinstance(t_tl, float):
            t_tl_s = t_tl

        data_single = tl_simulation_single(k, s, kb, t_int, t_tl_s, N[i])

        data = {**data, **data_single}

    return data



"""
Module with functions to plot survival functions.
"""

import pathlib
from typing import Dict, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np

from .. import compute, data_utils
from . import _plot_utils


# TODO: UPDATE DOCSTRING
def _base_data_multiple(
    data,
    process_data_flag: bool = True,
    xlim: Tuple[float, float] = None,
    ylim: Tuple[float, float] = None,
    figsize: Tuple[float, float] = (8, 5),
    kwargs_plot: Dict = None,
    kwargs_text: Dict = None,
):
    """_summary_

    Parameters
    ----------
    data : _type_
        _description_
    process_data_flag : bool, optional
        _description_, by default True
    xlim : Tuple[float, float], optional
        _description_, by default None
    ylim : Tuple[float, float], optional
        _description_, by default None
    figsize : Tuple[float, float], optional
        _description_, by default (8, 5)
    kwargs_plot : Dict, optional
        _description_, by default None
    kwargs_text : Dict, optional
        _description_, by default None

    Returns
    -------
    fig : matplotlib.Figure
        TODO: link this to matplotlib.Figure documentation
        https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure
    ax: matplotlib.axes.Axes
        TODO: link to the matplotlib.axes.Axes documentation
        https://matplotlib.org/stable/api/axes_api.html#matplotlib.axes.Axes
    """

    # Create empty dictionaries for the kwargs_... if they are None.
    # Mutable default values leads to dangerous behaviour
    # (https://stackoverflow.com/questions/1132941/least-astonishment-and-the-mutable-default-argument)
    # Hence, it is done this way.

    if kwargs_plot is None:
        kwargs_plot = dict()
    if kwargs_text is None:
        kwargs_text = dict()

    # Check if it is a single dict or a sequence
    # if it is a single dict, make it a sequence so it works with the rest of the
    # function
    if isinstance(data, dict):
        data = [data]

    if process_data_flag:
        # Process data
        for i in range(len(data)):
            data[i] = data_utils.process_data(data[i])

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)

    text_points = set()

    # Sort the time-lapse conditions from low to high, required for proper
    # text positions
    for i, d in enumerate(data):

        # Plotting kwargs
        kwargs_d = _plot_utils._get_key_to_value_i(i, kwargs_plot)

        for t_tl in sorted(data[i].keys()):
            time = d[t_tl]["time"]
            value = d[t_tl]["value"]

            ax.loglog(time, value, **kwargs_d)

            if i == 0:
                # Create the new point position for the text
                text_point = (time[0] * 0.6, value[0] * 1.1)

                # Check if the new text will not overlap
                if text_point in text_points:
                    text_point = (time[0] * 1.3, value[0] * 1.1)

                # This is a hack-job and should not be a long term fix
                # TODO: fix this do it automatically
                a = _plot_utils._fmt_t_str_plot(t_tl)
                if a == "$0.39\,\mathrm{s}$":
                    text_point = (time[0] * 1.1, value[0] * 1.1)

                ax.text(
                    text_point[0],
                    text_point[1],
                    _plot_utils._fmt_t_str_plot(t_tl),
                    **kwargs_text,
                )
                text_points.add(text_point)

    # Axis limits
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    # Labels
    ax.set_xlabel("time (s)")
    ax.set_ylabel("survival function")

    # Legend
    # Remove duplicate labels
    # Implementation from: https://stackoverflow.com/questions/13588920/stop-matplotlib-repeating-labels-in-legend
    handles, labels = ax.get_legend_handles_labels()
    by_label = dict(zip(labels, handles))
    ax.legend(by_label.values(), by_label.keys())  # , loc="lower left")

    return fig, ax


def data_multiple(
    data: Dict[str, Dict[str, np.ndarray]],
    process_data_flag: bool = True,
    xlim: Tuple[float, float] = None,
    ylim: Tuple[float, float] = None,
    figsize: Tuple[float, float] = (8, 5),
    kwargs_plot: Dict = None,
    kwargs_text: Dict = None,
):
    """Function plots a single or multiple data dictionaries"""

    # Create empty dictionaries for the kwargs_... if they are None.
    # Mutable default values leads to dangerous behaviour
    # (https://stackoverflow.com/questions/1132941/least-astonishment-and-the-mutable-default-argument)
    # Hence, it is done this way.

    if kwargs_plot is None:
        kwargs_plot = dict()
    if kwargs_text is None:
        kwargs_text = dict()

    # Set the default settings if they are not provided
    if "label" not in kwargs_plot:
        kwargs_plot["label"] = ["data1", "data2"]
    if "color" not in kwargs_plot:
        kwargs_plot["color"] = ["#007972", "#fe9901"]
    if "linewidth" not in kwargs_plot:
        kwargs_plot["linewidth"] = 1
    if "linestyle" not in kwargs_plot:
        kwargs_plot["linestyle"] = ["solid", "dashed"]

    fig, ax = _base_data_multiple(
        data,
        process_data_flag=process_data_flag,
        xlim=xlim,
        ylim=ylim,
        figsize=figsize,
        kwargs_plot=kwargs_plot,
        kwargs_text=kwargs_text,
    )

    return fig, ax


# TODO: fix bug in compute_multi_exp call
# TODO: UPDATE DOCSTRING
# TODO: check how .plot() function shows kwargs, it puts it under "Other parameters"
# although we know have it as a keyword argument with a default value, so we should
# likely just put it under "Parameters"
def data_vs_multi_exp(
    data: Dict[str, Dict[str, np.ndarray]],
    fit_values_multi_exp: Dict[str, Dict[str, np.ndarray]],
    process_data_flag: bool = True,
    xlim: Tuple[float, float] = None,
    ylim: Tuple[float, float] = None,
    figsize: Tuple[float, float] = (8, 5),
    kwargs_plot: Dict = None,
    kwargs_text: Dict = None,
):
    """Function plots the survival function of the true data and the multi-exponential curves.

    Parameters
    ----------
    data: Dict[str, Dict[str, np.ndarray]]
        Data of the survival function from real data with the following data structure:
        {
            "t_tl": {
                "time": np.ndarray with all the time values,
                "value": np.ndarray with all the survival function values corresponding to the
                    respective time value
            }
        }
    data_multi_exp: Dict[str, Dict[str, np.ndarray]]
        Survival function data for time-lapse conditions with the following data
        structure:
            {
                "t_tl": {
                    "time": np.ndarray with the time points,
                    "value": np.ndarray with the survival function values,
                },
                ...
            }
    """

    # Create empty dictionaries for the kwargs_... if they are None.
    # Mutable default values leads to dangerous behaviour
    # (https://stackoverflow.com/questions/1132941/least-astonishment-and-the-mutable-default-argument)
    # Hence, it is done this way.

    if kwargs_plot is None:
        kwargs_plot = dict()
    if kwargs_text is None:
        kwargs_text = dict()

    key = fit_values_multi_exp.keys()[0]
    n_exp = fit_values_multi_exp[key]["k"].shape[0]

    # Set the default settings if they are not provided
    if "label" not in kwargs_plot:
        kwargs_plot["label"] = ["data", f"{n_exp}-exp"]
    if "color" not in kwargs_plot:
        kwargs_plot["color"] = ["#007972", "#fe9901"]
    if "linewidth" not in kwargs_plot:
        kwargs_plot["linewidth"] = 1
    if "linestyle" not in kwargs_plot:
        kwargs_plot["linestyle"] = ["solid", "dashed"]

    data_multi_exp = compute.compute_multi_exp(fit_values_multi_exp, data)

    fig, ax = _base_data_multiple(
        [data, data_multi_exp],
        process_data_flag=process_data_flag,
        xlim=xlim,
        ylim=ylim,
        figsize=figsize,
        kwargs_plot=kwargs_plot,
        kwargs_text=kwargs_text,
    )

    return fig, ax


# TODO: UPDATE DOCSTRING


def data_vs_grid(
    data: Dict[str, Dict[str, np.ndarray]],
    fit_values_grid: Dict[str, Dict[str, np.ndarray]],
    process_data_flag: bool = True,
    xlim: Tuple[float, float] = None,
    ylim: Tuple[float, float] = None,
    figsize: Tuple[float, float] = (8, 5),
    kwargs_plot: Dict = None,
    kwargs_text: Dict = None,
):
    """Function plots the survival function of the true data and the GRID curves.

    Parameters
    ----------
    data: Dict[str, Dict[str, np.ndarray]]
        Survival function data for every time-lapse condition with the following data
        structure:
            {
                "t_tl": {
                    "time": np.ndarray with the time points,
                    "value": np.ndarray with the survival function values,
                }
            }
    xlim: Tuple[float, float] = None, ylim: Tuple[float, float] = None, path_save: str or pathlib.Path
        Path designates the place where the figure should be saved if a value is set. The figure is
        not saved if the values is set to None. (default None)

    Returns
    -------
    None
    """

    # Create empty dictionaries for the kwargs_... if they are None.
    # Mutable default values leads to dangerous behaviour
    # (https://stackoverflow.com/questions/1132941/least-astonishment-and-the-mutable-default-argument)
    # Hence, it is done this way.

    if kwargs_plot is None:
        kwargs_plot = dict()
    if kwargs_text is None:
        kwargs_text = dict()

    # Set the default settings if they are not provided
    if "label" not in kwargs_plot:
        kwargs_plot["label"] = ["data", "GRID"]
    if "color" not in kwargs_plot:
        kwargs_plot["color"] = ["#007972", "#fe9901"]
    if "linewidth" not in kwargs_plot:
        kwargs_plot["linewidth"] = 1
    if "linestyle" not in kwargs_plot:
        kwargs_plot["linestyle"] = ["solid", "dashed"]

    k = fit_values_grid["grid"]["k"]
    s = fit_values_grid["grid"]["s"]
    a = fit_values_grid["grid"]["a"]

    data_grid = compute.compute_grid_curve(k, s, a, data)

    fig, ax = _base_data_multiple(
        [data, data_grid],
        process_data_flag=process_data_flag,
        xlim=xlim,
        ylim=ylim,
        figsize=figsize,
        kwargs_plot=kwargs_plot,
        kwargs_text=kwargs_text,
    )

    return fig, ax



"""
Module for GRID and multi-exponential fitting.
"""

import math
from typing import Tuple

import numpy as np
import scipy.optimize

import gridlib.calc as calc
import gridlib.data_utils as data_utils


def create_fixed_grid(
    k_min: float, k_max: float, n: int, scale: str
) -> Tuple[np.ndarray, np.ndarray]:
    """Function creates the fixed grid based on the provided parameters.

    Parameters
    ----------
    k_min: float
        Minimum decay rate value.
    k_max: float
        Maximum decay rate value.
    n: int
        Number of positions in the grid.
    scale: str
        The scale to be applied for the creation of the grid.
        Options:
            "log": logarithmic scale
            "linear": linear scale

    Returns
    -------
    k: np.ndarray
        The fixed grid positions based on the given parameter values.
    """

    if scale not in {"log", "linear"}:
        raise ValueError("Scale should be either 'log' or 'linear'.")

    # Create the grid with fixed spacing between every position
    if scale == "linear":
        k = np.linspace(k_min, k_max, num=n, endpoint=True)
    elif scale == "log":
        k = np.geomspace(k_min, k_max, num=n, endpoint=True)

    return k


# Module for GRID fitting, multi-exponential fitting, and GRID resampling
def con_eq(x0: np.ndarray):
    """Constraint equality: makes sure all the amplitudes sum to 1."""
    s = x0[:-1]
    return np.sum(s) - 1


def fit_grid(parameters, data):

    data_processed = data_utils.process_data(data)

    if not data_utils.isvalid_parameters_grid(parameters):
        raise ValueError("parameters does not contain a valid value")

    if "k" in parameters:
        k = parameters["k"]
    else:
        k = create_fixed_grid(
            parameters["k_min"],
            parameters["k_max"],
            parameters["N"],
            parameters["scale"],
        )

    reg_weight = parameters["reg_weight"]

    x0 = np.ones(k.shape[0], dtype=np.float64) * (1 / k.shape[0])

    # Update the x0 array and set the bounds for both the amplitudes and the
    # photobleaching number
    lbq = 0  # lower bound photobleaching
    ubq = 3  # upper bound photobleaching

    bnds = [(0, 1) for _ in range(x0.shape[0])]
    bnds.append((lbq, ubq))  # add the bounds for the photobleaching
    if parameters["fit_a"]:

        tau_min = min(data_utils.get_time_sec(t_tl) for t_tl in data.keys())
        # The initial kb guess is 2 s^-1 as default
        a = 1.0 * tau_min  # # default bleaching rate is 1 s^-1
        x0 = np.concatenate((x0, np.array([a], dtype=np.float64)))

    elif not parameters["fit_a"]:
        a = parameters["a_fixed"]
        x0 = np.concatenate((x0, np.array([a])))
        bnds[-1] = (a, a)  # Fix the photobleaching number

    # print(x0.shape)
    # print(bnds)

    cons = [{"type": "eq", "fun": con_eq}]
    # TODO: change iprint, see source code
    options = {"maxiter": 600, "disp": True, "iprint": 1, "ftol": 10 ** (-11)}

    res = scipy.optimize.minimize(
        calc.lsqobj_grid,
        x0,
        args=(data_processed, k, reg_weight),
        method="SLSQP",
        jac=True,  # jac=True, means the gradient is given by the function
        bounds=bnds,
        constraints=cons,
        options=options,
    )

    grid_results = {
        "k": k,
        "s": res.x[: k.shape[0]],
        "a": res.x[-1],
        "loss": res.fun,
    }

    fit_results = {"grid": grid_results}

    return fit_results


def _fit_n_exp(parameters, data, n: int = 2):
    """Function to fit an n-exponential to the provided data and returns the fitted
    parameter results."""

    data_processed = data_utils.process_data(data)

    if not data_utils.isvalid_parameters_n_exp(parameters):
        raise ValueError("parameters does not contain a valid value")

    k_min = parameters["k_min"]
    k_max = parameters["k_max"]

    coordinates = []
    for i in range(n):
        coordinates.append(
            np.geomspace(k_min, k_max, num=2, endpoint=True, dtype=np.float64)
        )

    # Create the grid
    x0_k = np.meshgrid(*coordinates)
    x0_k = np.reshape(x0_k, (-1, n))

    # Initial guesses
    x0_s = np.linspace(0.05, 0.7, num=n, endpoint=True, dtype=np.float64)

    # Create the arrays for the bounds of the decay rates, and amplitude values
    # +1 for the photobleaching bounds
    lb = np.zeros(x0_k.shape[1] + x0_s.shape[0] + 1)
    ub = np.ones(x0_k.shape[1] + x0_s.shape[0] + 1)

    # Set the bounds for the decay rates
    lb[: x0_k.shape[1]] = k_min
    ub[: x0_k.shape[1]] = k_max

    # Bounds for the amplitudes are already set, since they are the
    # default values, see initialization of the lb and ub array
    # # Set the bounds for the amplitudes
    # lb[x0_k.shape[1] : (x0_k.shape[1] + x0_s.shape[0])] = 0
    # ub[x0_k.shape[1] : (x0_k.shape[1] + x0_s.shape[0])] = 1

    # Set the initial photobleaching number and the photobleaching bounds
    if parameters["fit_a"]:
        tau_min = min(data_utils.get_time_sec(t_tl) for t_tl in data.keys())
        # The initial kb guess is 2 s^-1 as default
        a = 1.0 * tau_min  # # default bleaching rate is 1 s^-1

        # Default values for photobleaching bounds
        lbq = 0  # lower bound photobleaching
        ubq = 3  # upper bound photobleaching

        lb[-1] = lbq
        ub[-1] = ubq
    elif not parameters["fit_a"]:
        a = parameters["a_fixed"]
        lb[-1] = parameters["a_fixed"]
        ub[-1] = parameters["a_fixed"]

    bnds = (lb, ub)

    print(f"Total fits: {x0_k.shape[0]}")

    fit_results_all = []

    for i in range(x0_k.shape[0]):
        x0 = np.concatenate((x0_k[i, :], x0_s, np.array([a])))
        res = scipy.optimize.least_squares(
            calc.global_multi_exp,
            x0,
            args=(data_processed, n),
            bounds=bnds,
            method="trf",
        )
        fit_results_all.append(res)

    # Find the best fitted parameters
    cost_min = math.inf
    x_best = None

    for res in fit_results_all:
        if res.cost < cost_min:
            cost_min = res.cost
            x_best = res.x

    # Retrieve the best decay rates
    k_temp = x_best[:n]

    # Retrieve the best weight values
    s_temp = x_best[n : (2 * n)]
    # Renormalize s
    s_norm = s_temp / np.sum(s_temp)

    # The indices to sort the k_array from low to high
    idx = np.argsort(k_temp)

    # Store the best results in a dictionary
    multi_exp_results = {
        "k": k_temp[idx],
        "s": s_norm[idx],
        "a": x_best[-1],
        "loss": cost_min,
    }

    # Store the results in a dictionary as a value with the key indicating the
    # number of decay rates
    fit_results = {f"{n}-exp": multi_exp_results}

    return fit_results


def fit_multi_exp(parameters, data):

    if not data_utils.isvalid_parameters_n_exp(parameters):
        raise ValueError("parameters does not contain a valid value")

    n = parameters["n_exp"]

    fit_results = dict()

    if isinstance(n, int):
        fit_result_n_exp = _fit_n_exp(parameters, data, n=n)
        fit_results = {**fit_results, **fit_result_n_exp}
    else:
        for n_value in n:
            fit_result_n_exp = _fit_n_exp(parameters, data, n=n_value)
            fit_results = {**fit_results, **fit_result_n_exp}

    return fit_results


def fit(parameters, data):
    """Function to fit both with GRID and with multi-exponentials."""

    parameters_grid, parameters_n_exp = data_utils.split_parameters(parameters)

    fit_results_grid = fit_grid(parameters_grid, data)
    fit_results_multi_exp = fit_multi_exp(parameters_n_exp, data)

    fit_results = {**fit_results_grid, **fit_results_multi_exp}

    return fit_results


def _fit_n_exp_fixed_k(parameters, data, n: int = 2):
    """Function to fit an n-exponential to the provided data and returns the fitted
    parameter results."""

    data_processed = data_utils.process_data(data)

    if not data_utils.isvalid_parameters(parameters):
        raise ValueError("parameters does not contain a valid value")

    if "k" in parameters:
        x0_k = np.copy(parameters["k"])
        x0_k = np.reshape(x0_k, (-1,))

        n = x0_k.shape[0]
    else:
        k_min = parameters["k_min"]
        k_max = parameters["k_max"]

        coordinates = []
        for i in range(n):
            coordinates.append(
                np.geomspace(k_min, k_max, num=2, endpoint=True, dtype=np.float64)
            )

        # Create the grid
        x0_k = np.meshgrid(*coordinates)
        x0_k = np.reshape(x0_k, (-1, n))

    # Create the arrays for the bounds of the decay rates, and amplitude values
    # +1 for the photobleaching bounds
    lb = np.zeros(x0_k.shape[1] + x0_s.shape[0] + 1)
    ub = np.ones(x0_k.shape[1] + x0_s.shape[0] + 1)

    # Initial guesses
    x0_s = np.linspace(0.05, 0.7, num=n, endpoint=True, dtype=np.float64)

    if "k" in parameters:
        # Set the bounds for the decay rates
        lb[: x0_k.shape[1]] = x0_k
        ub[: x0_k.shape[1]] = x0_k
    else:
        # Set the bounds for the decay rates
        lb[: x0_k.shape[1]] = k_min
        ub[: x0_k.shape[1]] = k_max

    # Bounds for the amplitudes are already set, since they are the
    # default values, see initialization of the lb and ub array
    # # Set the bounds for the amplitudes
    # lb[x0_k.shape[1] : (x0_k.shape[1] + x0_s.shape[0])] = 0
    # ub[x0_k.shape[1] : (x0_k.shape[1] + x0_s.shape[0])] = 1

    # Set the initial photobleaching number and the photobleaching bounds
    if parameters["fit_a"]:
        tau_min = min(data_utils.get_time_sec(t_tl) for t_tl in data.keys())
        # The initial kb guess is 2 s^-1 as default
        a = 1.0 * tau_min  # # default bleaching rate is 1 s^-1

        # Default values for photobleaching bounds
        lbq = 0  # lower bound photobleaching
        ubq = 3  # upper bound photobleaching

        lb[-1] = lbq
        ub[-1] = ubq
    elif not parameters["fit_a"]:
        a = parameters["a_fixed"]
        lb[-1] = parameters["a_fixed"]
        ub[-1] = parameters["a_fixed"]

    bnds = (lb, ub)

    print(f"Total fits: {x0_k.shape[0]}")

    fit_results_all = []

    for i in range(x0_k.shape[0]):
        x0 = np.concatenate((x0_k[i, :], x0_s, np.array([a])))
        res = scipy.optimize.least_squares(
            calc.global_multi_exp,
            x0,
            args=(data_processed, n),
            bounds=bnds,
            method="trf",
        )
        fit_results_all.append(res)

    # Find the best fitted parameters
    cost_min = math.inf
    x_best = None

    for res in fit_results_all:
        if res.cost < cost_min:
            cost_min = res.cost
            x_best = res.x

    # Retrieve the best decay rates
    k_temp = x_best[:n]

    # Retrieve the best weight values
    s_temp = x_best[n : (2 * n)]
    # Renormalize s
    s_norm = s_temp / np.sum(s_temp)

    # The indices to sort the k_array from low to high
    idx = np.argsort(k_temp)

    # Store the best results in a dictionary
    multi_exp_results = {
        "k": k_temp[idx],
        "s": s_norm[idx],
        "a": x_best[-1],
        "loss": cost_min,
    }

    # Store the results in a dictionary as a value with the key indicating the
    # number of decay rates
    fit_results = {f"{n}-exp": multi_exp_results}

    return fit_results
